#include <Eigen/Sparse>
#include <iostream>
#include "writer.hpp"
#include <cmath>
#include <Eigen/SparseCholesky>
#include <stdexcept>

//! Sparse Matrix type. Makes using this type easier.
typedef Eigen::SparseMatrix<double> SparseMatrix;

//! Used for filling the sparse matrix.
typedef Eigen::Triplet<double> Triplet;


//! Vector type
typedef Eigen::VectorXd Vector;


//! Our function pointer, typedef'd to make it easier to use
typedef double(*FunctionPointer)(double);


//! Create the Galerkin matrix for 1D with boundary terms
//! @param[out] A will be the Galerkin matrix (as in the exercise)
//! @param[in] N as in the exercise
//! @param[in] dx the cell length
void createGalerkinMatrixWithBoundary(SparseMatrix& A, int N, double dx) {
    // Fill in the matrix here, see the first exercise for how to do this.
}

//! Creates the right hand side for the finite element method in the exericse
//! @param[out] rhs the resulting right hand side (or phi in the exercise)
//! @param[in] g the function pointer to g
//! @param[in] dx the cell length
//! @param[in] x the x points
void createRHS(Vector& rhs, FunctionPointer g, int N, double dx, const Vector& x) {
    // Create RHS here according to the exercise
}


//! Solve the equation
//!
//!    \f[ -u''(x) = g(x)\f]
//!
//! using FEM on the interval \f$[a,b]\f$.
//!
//! @param[out] u at the end, will have all the values of u
//! @param[out] x will be the x points
//! @param[in] g should be a pointer to the function g
//! @param[in] N number of inner cells to use
//! @param[in] a endpoint on left side
//! @param[in] b endpoint on right side
//! @param[in] ua boundary value at a
//! @param[in] ub boundary value at b
void femSolve(Vector& u, Vector& x, FunctionPointer g, int N, double a, double b,
              double ua = 0.0, double ub = 0.0) {

    double dx = (b - a) / (N + 1);

    // Fill x vector
    x.setLinSpaced(N+2, a, b);


    SparseMatrix A;
    createGalerkinMatrixWithBoundary(A, N, dx);


    Vector phi;
    createRHS(phi, g, N, dx, x);

    // Solve the linear system.
    // for the first part exercise, you can ignore the boundary values.

}


void computeWithoutBoundaryValues() {
    Vector u;
    Vector x;
    femSolve(u, x, std::sin, 100, -M_PI, M_PI);

    writeToFile("u_wo_bc_fem.txt", u);
    writeToFile("x_wo_bc_fem.txt", x);
}

void computeWithBoundaryValues() {
    Vector u;
    Vector x;
    femSolve(u, x, std::cos, 100, -M_PI, M_PI, -1.0/2.0, -1.0/2.0);

    writeToFile("u_w_bc_fem.txt", u);
    writeToFile("x_w_bc_fem.txt", x);
}

int main(int, char**) {
    computeWithoutBoundaryValues();
    computeWithBoundaryValues();
}




